//This file may be redistributed and modified only under the terms of
//the GNU General Public License (See COPYING for details).
//Copyright (C) 2003 - 2007 Simon Goodall

package WFUT;

import java.net.*;
import java.io.*;
import java.util.*;
import java.util.zip.CRC32;

/**
 * This class is the outer loop for updating files. This class contains the
 * outer loop code to update a channel. It is implemented as a thread for the
 * GUI so the console area can be updated.
 */
class UpdateThread extends Thread {
	public int numFiles = 0;
	public int filenum = 0;
	public boolean done = false;
	private Object channels[] = null;
	public boolean running = true;

	public void setChannels(Object c[]) {
		channels = c;
	}

	public FileObject current = null;

	/**
	 * This function checks the OS name to see if it is a system that requires the
	 * executable flag via chmod to be set.
	 * 
	 * @return True if exectuable via chmod needs to be set
	 */
	public static boolean canExec() {
		// TODO also add in a check for OS-X
		final String os_name = System.getProperty("os.name");
		final boolean ret = os_name.equals("Linux");
		return ret;
	}

	public void run() {
		running = true;
		done = false;
		final String server = Options.getInstance().getServerRoot();

		for (int i = 0; i < channels.length; ++i) {
			if (!running) {
				break;
			}
			final String channel = channels[i].toString();
			updateChannel(server, channel, null);
		}
		done = true;
	}

	public void updateChannel(final String server, final String channel_name, final String reference_path) {
		String channel = channel_name;
		done = false;

		// Here we do lots of stuff to try and find a valid channel file. It could be a legacy file in channels/channel_name.xml or it could be a
		// new style wfut.xml. We also need to check to see if the channel was specified as . and determine the channel name from wfut.xml, or the
		// current directory name. This assumes that the current dir name is the same as the channel (it must be for this code to work!).
		// If we are using the old style channel file, then we need to strip the channel dir from each file entry as the later code automatically
		// adds this in again.
		// We also have a "prefix" allowing an arbitrary dir to be specified that contains another channel. This needs to be correctly adjusted.
		// However, it is likely that this is not the case.
		// Hopefully sometime in the future ( e.g. several months after sear 0.6.2) we can remove the legacy channel loading code and make this function
		// much simpler.

		try {

			String prefix = Options.getInstance().getPrefix();

			ChannelList localMap = null;
			// handle "." to mean that the current dir is within a WFUT channel and get the real
			// channel name from the wfut.xml file.
			if (channel.equals(".")) {
				try {
					final String loc1 = "wfut.xml";
					File xmlFile = new File(loc1);
					if (xmlFile.exists()) {
						localMap = FileIO.getLocalList(loc1);
						channel = localMap.getName();
					} else {
						errorCallback("No wfut.xml found in current directory dir");
						return;                                      
					}
					// Mess about with the prefix to setup the correct paths for this channel
					prefix = new File("..").getPath() + File.separator;

				} catch (FileNotFoundException e) {
					errorCallback("No wfut.xml found in current directory or legacy channels dir");
					return;
				}
			} else {
				try {
					final String loc1 = prefix + File.separator + channel + File.separator + "wfut.xml";

					File xmlFile = new File(loc1);
					if (xmlFile.exists()) {
						localMap = FileIO.getLocalList(loc1); 
					} else {
						localMap = new ChannelList();
					}          
				} catch (FileNotFoundException e) {
					localMap = new ChannelList();
				}
				localMap.setName(channel);
			}

			
			// See if there is a reference set of data in  a sytem location (read-only).
			ChannelList systemMap = null;
			// handle "." to mean that the current dir is within a WFUT channel and get the real
			// channel name from the wfut.xml file.
			if (reference_path == null) {
				systemMap = new ChannelList();
			} else {
				try {
					final String loc1 = reference_path + File.separator + "wfut.xml";

					File xmlFile = new File(loc1);
					if (xmlFile.exists()) {
						systemMap = FileIO.getLocalList(loc1); 
					} else {
						systemMap = new ChannelList();
					}          
				} catch (FileNotFoundException e) {
					systemMap = new ChannelList();
				}
			}

			// Setup base path for local files and base url for remote files
			final String basePath = prefix + File.separator + channel + File.separator; 
			final String baseURL = server + "/" + channel + "/"; 

			// Setup the path name for the temp download xml file
			final String tmpfile =  basePath + "tempwfut.xml";

			ChannelList serverMap = null;
			try {
				serverMap = FileIO.getChannelList(baseURL + "wfut.xml");
			} catch (FileNotFoundException e) {
				errorCallback("File not found: " + baseURL + "wfut.xml");
				System.err.println(e);
				return;
			}

			// Read the temp file in case of broken downloads!
			final File f = new File(basePath + "tempwfut.xml");
			if (f.exists()) {
				try {
					ChannelList tempMap = FileIO.getLocalList(f.toString());
					localMap.getFiles().putAll(tempMap.getFiles());
				} catch (FileNotFoundException e) {
					// assert (false);
					// Shouldn't get here from the f.exists() check
					errorCallback(e.toString());
					System.err.println(e);
				}
			}
			// Get list of files that need updating
			final java.util.List l = getUpdates(serverMap, localMap, systemMap, basePath);
			numFiles = l.size();
			filenum = 0;
			channelBegin(channel, numFiles);

			final Iterator itr = l.iterator();
			while (running && itr.hasNext()) {
				current = (FileObject) itr.next();
				++filenum;

				final String fname = current.getFilename();

				fileDownloading(fname);

				final String src = baseURL + fname;
				final String dst = basePath + fname;

				if (getFile(src, dst, current.getCRC32(), current.getExecute())) {
					// Add successful download to temp list
					fileDownloadComplete(fname);
					FileIO.saveTempList(tmpfile, channel, current);
					localMap.addFile(current);
				} else {
					this.fileDownloadError(fname);
					// Dont save broken files to list
					//serverMap.removeFile(current.getFilename());
				}
			}
			if (!running) {
				return;
			}
			current = null;
			try {
				System.out.println("Saving list file: " + basePath + "wfut.xml");
				FileIO.saveLocalList(localMap, basePath + "wfut.xml");
				// delete temp channel list
				f.delete();
			} catch (Exception e) {
				errorCallback("Error saving channel list");
				System.err.println(e);
			}
			channelComplete(channel);      
		} catch (Exception e) {
			System.err.println(e);
		}
	}

	public static String toURL(final String str) {
		return Encoder.encode(str);
	}

	public boolean getFile(final String src, final String dest, long crc32, boolean exec) throws IOException {    
		final int bufferSize = Options.getInstance().getBufferSizeInteger();

		// Create directory structure if required
		String parentDir = ".";
		if (dest.lastIndexOf(File.separator) > 0) {
			parentDir = dest.substring(0, dest.lastIndexOf(File.separator));
			File f = new File(parentDir);
			if (!f.exists()) {
				System.out.println("Making dirs: " + f.toString());
				if (f.mkdirs() == false) {
					// Error creating directory structure
					errorCallback("Error creating dirs: " + f.toString());
					return false;
				}
			}
		}
		boolean success = false;
		// Create a tempory file to download to first
		File f = new File(parentDir + File.separator + "tempFile");
		f.deleteOnExit(); // Remove on exit if it still exists

		BufferedInputStream in = null;
		BufferedOutputStream out = null;
		try {
			final URL url = new URL(toURL(src));
			final HttpURLConnection c = (HttpURLConnection) url.openConnection();
			final InputStream is = c.getInputStream();
			in = new BufferedInputStream(is);
			out = new BufferedOutputStream(new FileOutputStream(f));
			// Downloaded bytes are stored here
			byte[] buffer = new byte[bufferSize];
			int numberRead = 0;
			// Check CRC32 value
			CRC32 crc = new CRC32();
			while (running && (numberRead = in.read(buffer)) >= 0) {
				// Output bytes
				out.write(buffer, 0, numberRead);
				// Update crc value
				crc.update(buffer, 0, numberRead);
			}
			in.close();
			in = null;
			out.close();
			out = null;

			if (!running) {
				success = false;
			} else if (crc.getValue() != crc32) {

				errorCallback("CRC check failed");
				success = false;
			} else {
				// Rename temp file to real one if CRC is ok
				File destF = new File(dest);
				if (destF.exists()) {
					// What should we do if we can't delete the file?
					destF.delete();
				}
				// If it cant rename the file, then there are probs
				if (!f.renameTo(destF)) {
					errorCallback("Error renaming file");
					success = false;
				} else {
					// Only if we get here can we assume that we successfully downloaded the file
					success = true;
				}

				// Perform hacky bit to exec chmod to set files to be executable
				// if required.
				if (exec && canExec()) {
					try {
						Runtime.getRuntime().exec("chmod +x " + dest);
					} catch (IOException ioe) {
						// Ignore for now as this is really a *nix hack.
						// Users complain on windows that it's broken as chmod does not
						// exist
					}
				}
			}
		} catch (Exception e) {
			System.err.println(e);
			success = false;
		}
		// Close files before exiting
		finally {
			if (out != null) {
				out.close();
			}
			if (in != null) {
				in.close();
			}
		}

		return success;
	}


	public static java.util.List getUpdates(final ChannelList serverFiles, final ChannelList localFiles, final ChannelList systemFiles, final String ch_prefix) {
		final Map serverMap = serverFiles.getFiles();
		final Map localMap = localFiles.getFiles();
		final Map systemMap = systemFiles.getFiles();

		Iterator itr = serverMap.keySet().iterator();
		java.util.List l = new ArrayList();

		while (itr.hasNext()) {
			final String key = (String)itr.next();
			final FileObject server = (FileObject)serverMap.get(key);
			final FileObject local = (FileObject)localMap.get(key);
			final FileObject sys = (FileObject)systemMap.get(key);

			if (local == null) {
				if (sys == null) {
					// We need to download the file
					System.out.println("Added " + server.getFilename() + " as there is no local version");
					l.add(server);
				} else if (server.getVersion() > sys.getVersion()) {
					// Download into local repo
					l.add(server);
					System.out.println("Added " + server.getFilename() + " as sys version is older than server");
				} else {
					// Assume sys location will not be removed
				}

			} else if (server.getVersion() > local.getVersion()) {
				System.out.println("Added " + server.getFilename() + " as server is newer than local");
				// We need to overwrite the file
				l.add(server);
			} else {
				File f = new File(ch_prefix + File.separator + local.getFilename());
				// Check in case local file has been removed
				if (!f.exists()) {
					System.out.println("Added " + server.getFilename() + " as local version has been removed");
					l.add(server);
				} else {
					// Check CRC32 and print a warning if it differs.
					try {
						long crc32 = CRC32Calc.calcCRC32(f);
						if ( server.getCRC32() != crc32) {
							System.out.println(server.getFilename() + " has local modifications. Remove the file to update.");
						}
					} catch (IOException ioe) {
						System.err.println(ioe);
					}
				}
				// File is ok
			}
		}
		return l;
	}

	protected void fileDownloadError(final String filename) {
		Iterator i = m_callbacks.iterator();
		while (i.hasNext()) {
			WFUTCallbackHandler wc = (WFUTCallbackHandler)i.next();
			wc.DownloadFailed(filename);
		}

	}

	protected void fileDownloading(final String filename) {
		Iterator i = m_callbacks.iterator();
		while (i.hasNext()) {
			WFUTCallbackHandler wc = (WFUTCallbackHandler)i.next();
			wc.DownloadingFile(filename);
		}    
	}

	protected void fileDownloadComplete(final String filename) {
		Iterator i = m_callbacks.iterator();
		while (i.hasNext()) {
			WFUTCallbackHandler wc = (WFUTCallbackHandler)i.next();
			wc.DownloadComplete(filename);
		}
	}

	protected void channelBegin(final String channel, int numFiles) {
		Iterator i = m_callbacks.iterator();
		while (i.hasNext()) {
			WFUTCallbackHandler wc = (WFUTCallbackHandler)i.next();
			wc.ChannelBegin(channel, numFiles);
		}
	}

	protected void channelComplete(final String channel) {
		Iterator i = m_callbacks.iterator();
		while (i.hasNext()) {
			WFUTCallbackHandler wc = (WFUTCallbackHandler)i.next();
			wc.ChannelCompleted(channel);
		}
	}

	protected void errorCallback(final String msg) {
		Iterator i = m_callbacks.iterator();
		while (i.hasNext()) {
			WFUTCallbackHandler wc = (WFUTCallbackHandler)i.next();
			wc.Error(msg);
		}
	}

	private List m_callbacks = new ArrayList();

	public void addCallbackHandler(WFUTCallbackHandler wc) {
		m_callbacks.add(wc);   
	}

	public void removeCallbackHandler(WFUTCallbackHandler wc) {
		m_callbacks.remove(wc);
	}
}
